<?php

/**
 *  Server.php - enables quering UT2003 servers
 *  Copyright (c) 2002 Guillaume Cocatre-Zilgien <skamp@ut2003area.net>
 *  http://www.skamp.net/
 *
 *  This file is part of the UT2003area library.
 *
 *  The UT2003area library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  The UT2003area library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with the UT2003area library; if not, write to the
 *  Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Unreal is a registered trademark of Epic Games, Inc.
 *  Unreal (r) Tournament 2003 (c) 2002 Epic Games, Inc.
 */

class UT2003area_Server
{
    var $class_version;
    var $errors;
    
    var $server_address;
    var $server_port;
    var $query_port;
    var $timeout_seconds;
    var $timeout_microseconds;
    
    var $server_title;
    var $map_title;
    var $game_mode;
    var $num_players;
    var $max_players;
    
    var $server_vars;
    var $server_mode;
    var $admin_name;
    var $admin_email;
    var $goal_score;
    var $time_limit;
    var $min_players;
    
    var $player_ids;
    var $player_names;
    var $player_pings;
    var $player_scores;
    var $player_stats_ids;
    
    function UT2003area_Server($address, $port, $timeout = "0,300000")
    {
        $this->class_version    = "1.0RC1";
        $this->errors           = array ();
        
        $this->server_address   = $address;
        $this->server_port      = $port;
        $this->query_port       = $port + 1;
        
        $timeout_exploded = explode(",", $timeout);
        $this->timeout_seconds = (int)$timeout_exploded[0];
        $this->timeout_microseconds = (int)$timeout_exploded[1];
        
        $this->server_title     = "";
        $this->map_title        = "";
        $this->game_mode        = "";
        $this->num_players      = 0;
        $this->max_players      = 0;
        
        $this->server_vars      = array ();
        $this->server_mode      = "";
        $this->admin_name       = "";
        $this->admin_email      = "";
        $this->goal_score       = 0;
        $this->time_limit       = 0;
        $this->min_players      = 0;
        
        $this->player_ids       = array ();
        $this->player_names     = array ();
        $this->player_pings     = array ();
        $this->player_scores    = array ();
        $this->player_stats_ids = array ();
    }
    
    function getInfos()
    {
        $basic_infos = $this->getBasicInfos();
        $server_infos = $this->getServerInfos();

        if ($basic_infos === false || $server_infos === false) {
            return false;
        }
        
        $this->server_title = $basic_infos[3];
        $this->map_title = $basic_infos[4];
        $this->game_mode = $basic_infos[5];
        $this->num_players = $basic_infos[6];
        $this->max_players = $basic_infos[7];
        
        $n_infos = count($server_infos);
        for ($i = 0; $i < $n_infos; $i += 2) {
            $this->server_vars[$server_infos[$i]] = $server_infos[$i + 1];
        }
        
        return true;
    }
    
    function listPlayers($sort = "scores", $order = "descending")
    {
        $players_infos = $this->getPlayersInfos();
        if ($players_infos === false) return false;
        $p = count($players_infos);
        if ($p > 0) {
            $this->player_ids       = array ();
            $this->player_names     = array ();
            $this->player_pings     = array ();
            $this->player_scores    = array ();
            $this->player_stats_ids = array ();
        }
        for ($i = 0; $i < $p; $i += 5) {
            $this->player_ids[]         = $players_infos[$i];
            $this->player_names[]       = $players_infos[$i + 1];
            $this->player_pings[]       = $players_infos[$i + 2];
            $this->player_scores[]      = $players_infos[$i + 3];
            $this->player_stats_ids[]   = $players_infos[$i + 4];
        }
        
        $sort_array = array ();
        if ($sort == "ids") {
            $sort = $this->player_ids;
        } elseif ($sort == "names") {
            $sort = $this->player_names;
        } elseif ($sort == "pings") {
            $sort = $this->player_pings;
        } elseif ($sort == "stats_ids") {
            $sort = $this->player_stats_ids;
        } else {
            $sort = $this->player_scores;
        }
        
        if ($order == "ascending") {
            asort($sort, SORT_REGULAR);
        } else {
            arsort($sort, SORT_REGULAR);
        }
        reset($sort);
        
        $players = array ();
        foreach ($sort as $key => $value) {
            $players[] = array (
                "id"        => $this->player_ids[$key],
                "name"      => $this->player_names[$key],
                "ping"      => $this->player_pings[$key],
                "score"     => $this->player_scores[$key],
                "stats_id"  => $this->player_stats_ids[$key]
            );
        }
        
        return $players;
    }
    
    function getBasicInfos()
    {
        $query = "\x78\x00\x00\x00\x00";
        $format = "H10;v1;H12;s3;V2;H12";
        $data = $this->_getServerData($query);
        if ($data === false) return false;
        $extracted = $this->_extractData($format, $data, 5);
        return $extracted;
    }
    
    function getServerInfos()
    {
        $query = "\x78\x00\x00\x00\x01";
        $format = "s0";
        $data = $this->_getServerData($query);
        if ($data === false) return false;
        $extracted = $this->_extractData($format, $data, 5);
        return $extracted;
    }

    function getPlayersInfos()
    {
        $query = "\x78\x00\x00\x00\x02";
        $format = "V1;s1;V3";
        $data = $this->_getServerData($query);
        if ($data === false) return false;
        $extracted = $this->_extractData($format, $data, 5, 0);
        return $extracted;
    }

    function _getServerData($query, $output = "bin")
    {
        $data = "";
        $query_port = $this->server_port + 1;
    
        $fp = @fsockopen("udp://" . $this->server_address, $this->query_port,
                $errno, $errstr);
        if (!$fp) {
            $this->errors[] = $errstr;
            return false;
        }
        socket_set_blocking($fp, true);
        socket_set_timeout($fp, $this->timeout_seconds,
                $this->timeout_microseconds);
        $fwrite = @fwrite($fp, $query);
        if ($fwrite == -1) {
            $fclose = @fclose($fp);
            $this->errors[] = "couldn't write data on opened socket";
            return false;
        }
        $continue = true;
        while ($continue) {
            $data .= @fgetc($fp);
            $status = socket_get_status($fp);
            $continue = $status["unread_bytes"];
        }
        $fclose = @fclose($fp);
        if (empty($data)) {
            $this->errors[] = "no data";
            return false;
        } else {
            $hex_query = bin2hex(substr($data, 0, 4));
            if ($hex_query != "78000000") {
                $this->errors[] = "server's answer to query is invalid";
                return false;
            }
        }
        if ($output == "hex") {
            return bin2hex($data);
        };
        return $data;
    }
    
    function _testEndOfLoop($position, $length, $i, $repeat)
    {
        if ($repeat > 0 && $i < $repeat) return true;
        if ($repeat == 0 && $position < $length) return true;
        return false;
    }
    
    function _extractData($infos_format, $data, $position = 0, $repeat = 1)
    {
        $extracted = array ();
        $data_length = strlen($data);
        $i = 0;
    
        // process format
        $format_exploded = explode(";", $infos_format);
        $formats = array ();
        $f_repeats = array ();
        foreach ($format_exploded as $fstring) {
            $formats[] = $fstring{0};
            $f_repeats[] = (int)substr($fstring, 1);
        }
        $n_formats = count($formats);
    
        while ($this->_testEndOfLoop($position, $data_length, $i, $repeat)) {
            for ($f = 0; $f < $n_formats; $f++) {
                $format = $formats[$f];
                $f_repeat = $f_repeats[$f];
                $r = 0;
                
                while ($this->_testEndOfLoop($position, $data_length, $r, $f_repeat)) {
                    if ($format == "H") {
                        $n_chars = $f_repeat / 2;
                        $bin_chunk = substr($data, $position, $n_chars);
                        $unpacked = unpack("H" . $f_repeat . "hex", $bin_chunk);
                        $extracted[] = $unpacked["hex"];
                        $position += $n_chars;
                        $r = $f_repeat;
                    } elseif ($format == "v") {
                        $n_chars = 2;
                        $bin_chunk = substr($data, $position, $n_chars);
                        $unpacked = unpack("vshort", $bin_chunk);
                        $extracted[] = (int)$unpacked["short"];
                        $position += $n_chars;
                        ++$r;
                    } elseif ($format == "V") {
                        $n_chars = 4;
                        $bin_chunk = substr($data, $position, $n_chars);
                        $unpacked = unpack("Vlong", $bin_chunk);
                        $extracted[] = (int)$unpacked["long"];
                        $position += $n_chars;
                        ++$r;
                    } elseif ($format == "s") {
                        $char = $data{$position};
                        $l = ord($char) - 1;
                        ++$position;
                        if ($l > 0) {
                            $extracted[] = substr($data, $position, $l);
                        } else {
                            $extracted[] = "";
                        }
                        $position += $l + 1; // don't take NULL char at the end
                        ++$r;
                    }
                }
            }
            ++$i;
        }
        return $extracted;
    }
}
 
?>
